﻿; Definitions
.include "System.inc"

.org $00
	di
	im 1
	jp Boot

.org $38
	reti

.org $66
	retn

; Libraries:
.include "Video.asm"

Boot:
	; Make sure SP points somewhere sensible.
	ld sp,Memory.Stack
	
	; Set paging to something sensible.
	xor a
	ld ($FFFC),a
	ld ($FFFD),a
	inc a
	ld ($FFFE),a
	inc a
	ld ($FFFF),a

Main:
	
	; Reset the VDP to sensible defaults.
	call Video.Reset
	
	; Clear VRAM and palette.
	call Video.ClearAll
	
	; Load the font.
	ld hl,FontTileIndex*8
	call Video.GotoHL
	
	ld hl,Font
	ld d,(Font.End-Font)/8
LoadChar:
	ld c,8
LoadCharRow:
	ld a,(hl)
	inc hl
	ld b,4
-:	out (Ports.Video.Data),a
	djnz -
	dec c
	jr nz,LoadCharRow
	dec d
	jr nz,LoadChar
	
	
	; Load the sprites
	ld hl,8*4*256
	call Video.GotoHL
	
	ld hl,Sprites
	ld d,(Sprites.End-Sprites)/9
LoadSprite:
	ld e,(hl) ; bitplanes
	inc hl
	ld c,8
LoadSpriteRow:
	push de
	ld a,(hl)
	ld b,4
-:	ld a,(hl)
	srl e
	jr c,+
	xor a
+:	out (Ports.Video.Data),a
	djnz -
	inc hl
	
	pop de
	dec c
	jr nz,LoadSpriteRow
	dec d
	jr nz,LoadSprite
	
	call LoadPalette
	
	xor a
	ld (CurRow),a
	ld (CurCol),a
	
	ld hl,HelloWorld
	call PutString
	
	
	; Clear sprites
	call StartSprites
	call EndSprites
	
	
	call Video.ScreenOn
	
	ld a,18
	ld (GunA.ScaleXMin),a
	ld (GunB.ScaleXMin),a
	ld a,145
	ld (GunA.ScaleXMax),a
	ld (GunB.ScaleXMax),a
	
GunLoop:

	; Check for reset button
	in a,($DD)
	bit 4,a
	jp z,$0000

	; Alternate between the two guns.
	call SampleOtherGun
	
	; Reset sample count and min/max coordinates.
	ld ix,(GunToSample)
	ld d,0
	
WaitNextGunSignal:
	ld a,(GunToSampleTHMask)
	ld b,a
WaitGunSignal:
	in a,($BF)
	or a
	jp m,FinishedGunSample ; vblank, we're done here.
	
	in a,($DD)
	and b
	jr nz,WaitGunSignal
	
	; We have a signal from the gun.
	in a,($7F) ; horizontal counter
	ld b,a
	
	in a,($7E) ; vertical counter
	add a,32
	ld c,a
	
	ld a,d
	or a
	jr nz,+
	
	inc d
	ld (ix+0),b
	ld (ix+1),b
	ld (ix+2),c
	ld (ix+3),c
	jr WaitNextGunSignal
	
+:

	; Count samples
	inc d
	jr nz,+
	dec d
+:

	ld a,b
	ld l,(ix+0)
	ld h,(ix+1)
	call MinMax
	ld (ix+0),l
	ld (ix+1),h
	
	ld a,c
	ld l,(ix+2)
	ld h,(ix+3)
	call MinMax
	ld (ix+2),l
	ld (ix+3),h
	
	jr WaitNextGunSignal
	

FinishedGunSample:
	
	; Store sample count.
	ld (ix+4),d
	
	; Clear sprites
	call StartSprites
	
	; Process and clear samples
	call ProcessSamples
	
	ld ix,GunA
	ld b,4
	call DisplaySamples
	
	ld ix,GunB
	ld b,5
	call DisplaySamples

	; End sprites
	call EndSprites
	
	
	jp GunLoop

MinMax:
	cp l
	jr nc,+
	ld l,a
+:	cp h
	ret c
	ld h,a
	ret
	
ProcessSamples:
	ld a,(ix+4)
	cp 4
	jr nc,FoundSamples
	
	ld a,$FF
	ld (ix+4),$FF ; OnScreen
	ret
	
	
FoundSamples:

	ld (ix+4),0 ; OnScreen
	
	ld a,(ix+0)
	call ScaleX
	ld (ix+0),a
	
	ld a,(ix+1)
	call ScaleX
	ld (ix+1),a
	
	ld a,(ix+2) ; MinY
	sub 32
	ld (ix+2),a
	ld b,a
	ld a,(ix+3) ; MaxY
	sub 32
	ld (ix+3),a
	sub b
	sra a
	ld b,a
	ld a,(ix+2) ; MinY
	add a,b
	ld (ix+6),a ; Y
	
	ld a,(ix+0) ; MinX
	add a,b
	ld (ix+5),a ; X
	
	ret
	
DisplaySamples:
	ld a,(ix+4)
	or a
	ret nz
	
	push bc
	
	ld a,(ix+0)
	ld b,a
	ld a,(ix+2)
	ld c,a
	ld a,0
	call PutSprite
	
	ld a,(ix+0)
	ld b,a
	ld a,(ix+3)
	dec a
	ld c,a
	ld a,1
	call PutSprite
	
	ld a,(ix+1)
	dec a
	ld b,a
	ld a,(ix+2)
	ld c,a
	ld a,2
	call PutSprite
	
	ld a,(ix+1)
	dec a
	ld b,a
	ld a,(ix+3)
	dec a
	ld c,a
	ld a,3
	call PutSprite

	; Crosshair

	ld a,(ix+5)
	sub 3
	ld b,a
	
	ld a,(ix+6)
	sub 3
	ld c,a
	
	pop af
	call PutSprite
	
	ret

PutCoords:
	push bc
	ld a,'('
	call PutChar
	
	pop bc
	push bc
	ld a,b
	call PutHexByte
	
	ld a,','
	call PutChar
	
	pop bc
	ld a,c
	call PutHexByte
	ld a,')'
	jp PutChar
	

FontTileIndex = 0
FontCharOffset = FontTileIndex-' '

.var ubyte CurRow, CurCol

PutMap:
	push af
	ld a,(CurRow)
	ld l,a
	ld h,0
	ld b,6
-:	add hl,hl
	djnz -
	ld a,(CurCol)
	add a,a
	ld e,a
	ld d,0
	add hl,de
	ld de,$3800
	add hl,de
	call Video.GotoHL
	pop af
	add a,FontCharOffset
	out (Ports.Video.Data),a
	xor a
	out (Ports.Video.Data),a
	ret
	
PutChar:
	cp '\r'
	jr nz,+
	xor a
	ld (CurCol),a
	ret
+:	cp '\n'
	jr nz,+
	xor a
	ld (CurCol),a
	jr NewLine
+:	call PutMap
	ld a,(CurCol)
	inc a
	cp 32
	jr nz,+
	xor a
+:	ld (CurCol),a
	ret nz
NewLine:
	ld a,(CurRow)
	inc a
	cp 24
	jr nz,+
	xor a
+:	ld (CurRow),a
	ret

PutString:
	ld a,(hl)
	inc hl
	or a
	ret z
	push hl
	call PutChar
	pop hl
	jr PutString

HelloWorld:
	;   "12345678901234567890123456789012"
	.db "\n\n\n"
	.db "   Hello, world!\n"
	.db "   =============\n\n"
	
	.db "   This is running on a\n"
	.db "   Sega Master System.\n"
	.db "\n"
	.db "                      Hooray!"
	.db "\n"
	.db 0

OffScreen:
	.db "Off screen!      ",0

PutHexNybble:
	cp 10
	jr c,+
	add a,'A'-10
	jp PutChar
+:	add a,'0'
	jp PutChar

PutHexByte:
	push af
	srl a
	srl a
	srl a
	srl a
	call PutHexNybble
	pop af
	and %1111
	jr PutHexNybble

PutHexWord:
	push hl
	ld a,h
	call PutHexByte
	pop hl
	ld a,l
	jr PutHexByte


Font:
.include "bbc"
Font.End:

LoadPalette:
	xor a
	call Video.GotoPalette
	ld hl,Palette
	ld b,32
-:	ld a,(hl)
	inc hl
	out (Ports.Video.Data),a
	djnz -
	ret

SetSprite:
	; bc = coords
	; a = sprite
	; l = index
	
	push af
	push bc
	
	ld h,$3F
	push hl
	call Video.GotoHL
	pop hl
	
	pop bc
	push bc
	
	ld a,c
	dec a
	out (Ports.Video.Data),a
	
	cp $D0
	jr nz,+
	
	pop bc
	pop af
	ret
	
	
+:
	
	ld a,l
	add a,a
	add a,$80
	ld l,a
	call Video.GotoHL
	
	pop bc
	ld a,b
	out (Ports.Video.Data),a
	
	pop af
	out (Ports.Video.Data),a
	ret

StartSprites:
	xor a
	ld (SpriteIndex),a
	ret

EndSprites:
	ld c,$D1

PutSprite:
	push af
	ld a,(SpriteIndex)
	ld l,a
	inc a
	ld (SpriteIndex),a
	pop af
	jp SetSprite


.var ubyte SpriteIndex


Palette:
.db %111001, %000000, %000000, %000000, %000000, %000000, %000000, %000000 ; Tiles   0..7
.db %000000, %000000, %000000, %000000, %000000, %000000, %000000, %111111 ; Tiles   8..F
.db %000000, %011101, %000011, %110000, %000000, %000000, %000000, %000000 ; Sprites 0..7
.db %000000, %000000, %000000, %000000, %000000, %000000, %000000, %110100 ; Sprites 8..F

Sprites:
.db %0001
.db %11000000 ; TL
.db %10000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000

.db %0001
.db %10000000 ; BL
.db %11000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000

.db %0001
.db %11000000 ; TR
.db %01000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000

.db %0001
.db %01000000 ; BR
.db %11000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000
.db %00000000

.db %0010
.db %00010000 ; Red Crosshair
.db %00010000
.db %00010000
.db %11111110
.db %00010000
.db %00010000
.db %00010000
.db %00000000

.db %0011
.db %00010000 ; Blue Crosshair
.db %00010000
.db %00010000
.db %11111110
.db %00010000
.db %00010000
.db %00010000
.db %00000000
Sprites.End:

.struct Gun
	ubyte XMin, ; 0
	ubyte XMax, ; 1
	ubyte YMin, ; 2
	ubyte YMax, ; 3
	ubyte OnScreen, ; 4
	ubyte X, ; 5
	ubyte Y, ; 6
	ubyte ScaleXMin, ; 7
	ubyte ScaleXMax ; 8

.var Gun GunA, GunB

.var uword GunToSample
.var ubyte GunToSampleTHMask

SampleGunA:
	ld hl,GunA
	ld (GunToSample),hl
	ld a,%01000000
	ld (GunToSampleTHMask),a
	ld a,%11110111
	out ($3F),a
	ret

SampleGunB:
	ld hl,GunB
	ld (GunToSample),hl
	ld a,%10000000
	ld (GunToSampleTHMask),a
	ld a,%11111101
	out ($3F),a
	ret

SampleOtherGun:
	ld a,(GunToSampleTHMask)
	bit 6,a
	jr z,SampleGunA
	jr SampleGunB

ScaleX:
	sub (ix+7) ; ScaleXMin
	jr nc,+
	xor a
+:
	ld h,a
	ld l,0
	ld a,(ix+8) ; ScaleXMax
	sub (ix+7) ; ScaleXMin
	
	ld b,l ; 0
	ld c,h
	or a
	sbc hl,bc
	
	ld c,a
	
	ld b,16
	xor a
-:	add hl,hl
	rla
	cp c
	jr c,$+4
	sub c
	inc l
	djnz -
	
	ld a,l
	ret

OffsetY:
	add a,(ix+9)
	ret